__version__ = "0.6.2"

import logging

from .benchmarking_utilities import Stopwatch
from .collections_utilities import in_batches
from .datetime_utilities import ensure_tzinfo
from .error_utilities import ExceptionsAsErrors, add_error_to
from .file_utilities import (
    abspath_if_relative,
    file_checksum,
    move_and_create_dest,
    silent_create_dirs,
    silent_remove,
    switch_extension,
)
from .logging_utilities import (
    SingleLineColoredFormatter,
    SingleLineFormatter,
    SQLFilter,
    StandardMetadataFilter,
    log_to_console_for,
    log_to_tmp_file_for,
    silence_logger,
)
from .metaprogramming_helpers import (
    all_subclasses,
    getval,
    import_string,
    leaf_subclasses,
)
from .os_utilities import current_user, current_user_home
from .string_utilities import is_blank

logging.getLogger(__name__).addHandler(logging.NullHandler())



import timeit


class Stopwatch(object):
    """
    Simple stopwatch that measures duration of block of code in [ms]

    Example:
        >>> import time
        >>>
        >>> with Stopwatch() as stopwatch:
        ...    time.sleep(1)
        >>> assert stopwatch.duration_ms >= 1000
    """

    def __init__(self):
        self.start = 0
        self.end = 0

    def __enter__(self):
        self.start = timeit.default_timer()
        return self

    def __exit__(self, *args):
        self.end = timeit.default_timer()
        return False

    @property
    def duration_ms(self):
        return (self.end - self.start) * 1000



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# seveno_pyutil documentation build configuration file, created by
# sphinx-quickstart on Tue Sep 26 10:26:35 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import shlex
import sys
from datetime import datetime
import sphinx


# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "src")))

nitpicky = True

# -- ReadTheDocs stuff  ------------------------------------------------
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "m2r2",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = [".rst"]

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "seveno_pyutil"
author = "Tomislav Adamic"
copyright = (
    ", ".join(str(y) for y in range(2017, datetime.now().year + 1)) + ", " + author
)

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
version = release = "0.6.2"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False


# -- Options for HTML output ----------------------------------------------
if not on_rtd:  # only import and set the theme if we're building docs locally
    import sphinx_rtd_theme

    html_theme = "sphinx_rtd_theme"
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]


# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "seveno_pyutildoc"


# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "seveno_pyutil.tex",
        "seveno\\_pyutil Documentation",
        "Tomislav Adamic",
        "manual",
    )
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "seveno_pyutil", "seveno_pyutil Documentation", [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "seveno_pyutil",
        "seveno_pyutil Documentation",
        author,
        "seveno_pyutil",
        "One line description of project.",
        "Miscellaneous",
    )
]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3.6", None),
    "marshmallow": ("https://marshmallow.readthedocs.io/en/3.0/", None),
    "json": ("https://simplejson.readthedocs.io/en/latest/", None),
    "simplejson": ("https://simplejson.readthedocs.io/en/latest/", None),
}


add_module_names = False



from seveno_pyutil.error_utilities import ExceptionsAsErrors, add_error_to


class DescribeExceptionsAsErrors:
    def it_handles_basic_cases(self):
        errors = {}

        with ExceptionsAsErrors(errors) as e:
            raise RuntimeError("ZOMG!")
        errors == {"_schema": ["ZOMG!"]}

        errors = {}
        with ExceptionsAsErrors(errors, subkey="some_name") as e:
            raise RuntimeError("ZOMG!")
        errors == {"some_name": ["ZOMG!"]}

    def it_handles_example_cases(self):
        errors = {
            "person": {
                "email": ["is not an email"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
            },
            "job": ["is not from allowed values list"],
        }

        add_error_to(errors, {"person": {"email": "is from illegal domain"}})

        assert errors == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
            },
            "job": ["is not from allowed values list"],
        }

        add_error_to(errors, {"person": "is illegally formed"})

        assert errors == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
                "_schema": ["is illegally formed"],
            },
            "job": ["is not from allowed values list"],
        }

        add_error_to(errors, {"job": {"title": "can't be blank"}})

        assert errors == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
                "_schema": ["is illegally formed"],
            },
            "job": {
                "_schema": ["is not from allowed values list"],
                "title": ["can't be blank"],
            },
        }

        add_error_to(errors, {"job": {"title": "is overpaid"}})
        add_error_to(errors, {"job": {"title": ["is forbiden", "doesn't exist"]}})

        assert errors == {
            "person": {
                "email": ["is not an email", "is from illegal domain"],
                "name": ["is too long"],
                "date_of_birth": ["is not a date"],
                "_schema": ["is illegally formed"],
            },
            "job": {
                "_schema": ["is not from allowed values list"],
                "title": [
                    "can't be blank",
                    "is overpaid",
                    "is forbiden",
                    "doesn't exist",
                ],
            },
        }

    def it_handles_deeply_nested_case(self):
        errors = {"foo": ["is not bar"]}
        add_error_to(errors, {"foo": {"bar": {"baz": "ZOMG"}}})
        assert errors == {"foo": {"_schema": ["is not bar"], "bar": {"baz": ["ZOMG"]}}}



import logging
import sys
import traceback
from logging import NullHandler


def silence_logger(logger):
    """
    For given logger, replaces all its handlers with
    :class:`logging.NullHandler`.
    """
    while logger.handlers:
        logger.removeHandler(logger.handlers[0])

    logger.addHandler(NullHandler())


def log_to_console_for(logger_name):
    """
    Sometimes, usually during development, we want to quickly see output of
    some particular package logger. This method will configure such logger
    to spit stuff out to console.
    """
    logger = logging.getLogger(logger_name)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)


def log_to_tmp_file_for(logger_name, file_path="/tmp/seveno_pyutil.log"):
    """
    Quick setup for given logger directing it to ``/tmp/seveno_pyutil.log``
    This is of course mainly used during development, especially when playing
    with things in Python console.
    """
    logger = logging.getLogger(logger_name)
    handler = logging.FileHandler(filename=file_path)
    handler.setLevel(logging.DEBUG)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(handler)



from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

from seveno_pyutil import ensure_tzinfo


class describe_ensure_tzinfo(object):
    def it_converts_naive_datetime_to_timezone_aware_datetime(self):
        dt = datetime.now()
        assert ensure_tzinfo(dt, tz_or_offset=None).tzinfo is ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="Z").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="UTC").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset=ZoneInfo("UTC")).tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="-02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02").tzinfo is not None
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=2, minutes=30)).tzinfo
            is not None
        )
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=-2, minutes=30)).tzinfo
            is not None
        )

    def it_converts_timezone_aware_datetime_to_different_timezone(self):
        dt = datetime.now().replace(tzinfo=ZoneInfo("Europe/Zagreb"))

        assert ensure_tzinfo(dt, tz_or_offset=None).tzinfo is ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="Z").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="UTC").tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset=ZoneInfo("UTC")).tzinfo == ZoneInfo("UTC")
        assert ensure_tzinfo(dt, tz_or_offset="-02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02:00").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+0200").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="-02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="02").tzinfo is not None
        assert ensure_tzinfo(dt, tz_or_offset="+02").tzinfo is not None
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=2, minutes=30)).tzinfo
            is not None
        )
        assert (
            ensure_tzinfo(dt, tz_or_offset=timedelta(hours=-2, minutes=30)).tzinfo
            is not None
        )



from datetime import date, datetime
from typing import AnyStr, Iterable, Union

try:
    import simplejson as json
except Exception:
    import json


def is_blank(obj: Union[AnyStr, Iterable, None]) -> bool:
    """
    True if obj is empty string, None, string that contains only spaces and space like
    characters, or iterable that contains only these kinds of strings/objects
    """
    if not obj:
        return True

    if isinstance(obj, (str, bytes)):
        return not obj.strip()

    retv = False
    try:
        retv = all(is_blank(x) for x in obj)
    except TypeError:
        # ... not iterable
        retv = False

    return retv



import re
from datetime import datetime, timedelta, tzinfo
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from dateutil.tz import tzoffset

_ISO_8601_OFFSET = re.compile(r"([+-]?)([0-9]{2})[:]?([0-9]{0,2})")


def timezone_or_offset(from_):
    """
    Given ``from_`` creates `datetime.tzinfo` or `dateutil.tz.tzoffset` as
    result.

    ``from_`` can be any of following:

    - `str` (ie. "-02:42", None, "", "Z", ...) which is ISO8601 offset
    - `str` (ie. "Europe/Zagreb") which is timezone name
    - `int` (ie. -9000) which is total number of seconds in time offset
    - `datetime.timedelta`
    - `datetime.tzinfo` or something that behaves like it
    """
    offset_obj = None

    if from_ is None:
        offset_obj = ZoneInfo("UTC")

    elif isinstance(from_, str):
        try:
            offset_obj = ZoneInfo(from_)
        except (ZoneInfoNotFoundError, ValueError):
            pass

        if not offset_obj:
            if from_.strip() in ["Z", ""]:
                offset_obj = ZoneInfo("UTC")
            else:
                match_object = _ISO_8601_OFFSET.match(from_)
                if match_object:
                    sign, hours, minutes = match_object.groups()
                    offset_obj = tzoffset(
                        name=from_,
                        offset=timedelta(
                            hours=(-1 if sign == "-" else 1) * int(hours),
                            minutes=int(minutes or 0),
                        ),
                    )

    elif isinstance(from_, timedelta):
        offset_obj = tzoffset(name="{} s".format(from_.total_seconds()), offset=from_)

    elif isinstance(from_, int):
        offset_obj = tzoffset(
            name="{} s".format(from_), offset=timedelta(seconds=from_)
        )

    elif issubclass(type(from_), tzinfo) or all(
        hasattr(from_, attr_name)
        for attr_name in ["dst", "fromutc", "tzname", "utcoffset"]
    ):
        offset_obj = from_

    if offset_obj is None:
        raise ValueError("Unable to parse time offset: {}".format(from_))

    return offset_obj


def ensure_tzinfo(val, tz_or_offset="UTC", is_dst=False):
    """
    Creates timezone aware datetime object for ``val``.

    - if ``val`` is naive datetime, new value will be created as datetime
      localized in ``tz_or_offset`` timezone
    - if ``val`` is already timezone aware, it will be converted to
      ``tz_or_offset`` timezone using `datetime.datetime.astimezone`

    Arguments:
        val(datetime.datetime): Input value for conversion
        tz_or_offset: Anything that `timezone_or_offset` accepts
        is_dst(bool): used to determine the correct timezone in the ambiguous
            period at the end of daylight saving time. Use ``is_dst=None`` to
            raise an AmbiguousTimeError for ambiguous times at the end of
            daylight saving time.

    Return:
        datetime.datetime: Timezone aware datetime object

    Raises:
        ValueError: When timezone of offset can't be parsed / determined from
            ``tz_or_offset``

    Note:
        This tries to provide safe(ish) implementation for handling naive
        datetime objects, but ultimate solution is to not use naive datetime
        objects ever/anywhere. Recommendation is to go with ``pip install
        pendulum`` and leave this crap behind to history.
    """
    if not isinstance(val, datetime):
        raise ValueError(
            "Input is not datetime! ensure_tzinfo doesn't parse datetime, you "
            "need to do that beforehand."
        )

    tz_or_offset = timezone_or_offset(tz_or_offset)

    if not val.tzinfo:
        val = val.replace(tzinfo=tz_or_offset)
    else:
        val = val.astimezone(tz_or_offset)

    return val



from .single_line_formatter import SingleLineColoredFormatter, SingleLineFormatter
from .sql_filter import SQLFilter
from .standard_metadata_filter import StandardMetadataFilter
from .utilities import log_to_console_for, log_to_tmp_file_for, silence_logger



from seveno_pyutil.metaprogramming_helpers import (
    all_subclasses,
    import_string,
    leaf_subclasses,
)


class Base(object):
    pass


class DirectChild1(Base):
    pass


class DirectChild2(Base):
    pass


class Child(DirectChild1):
    pass


def test_leaf_subclasses_returns_correct_data():
    assert set(leaf_subclasses(Base)) == set([DirectChild2, Child])


def test_all_subclasses_returns_correct_data():
    assert all_subclasses(Base) == set([DirectChild1, DirectChild2, Child])


def test_import_string_imports_class_from_string():
    klass = import_string("seveno_pyutil.benchmarking_utilities.Stopwatch")
    assert klass.__name__ == "Stopwatch"
    assert klass.__module__ == "seveno_pyutil.benchmarking_utilities"



import enum
import logging
import timeit
from datetime import date, datetime, timedelta

import pygments
import sqlparse

# from pygments.formatters import TerminalTrueColorFormatter
from pygments.formatters import Terminal256Formatter
from pygments.lexers import SqlLexer

try:
    import simplejson as json
except Exception:
    import json


class JsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (date, datetime)):
            return o.isoformat()
        if isinstance(o, enum.Enum):
            return o.name
        return super().default(o)


class SQLFilter(logging.Filter):
    """
    Filter for Django's and SQLAlchemy SQL loggers. Reformats and colorizes
    queries.

    To use it with Django:

    - add it to ``django.db.backends`` and ``django.db.backends.schema``
      loggers
    - remove ``%(message)s`` from log format because it will cause to double
      emit each SQL statement. Use this filter's placeholder's instead.

    To use it with SQLAlchemy:

    - add it to ``sqlalchemy.engine`` logger and call
      `:meth:register_sqlalchemy_logging_events`

    Supports following loging placeholders:

    +-------------------+----------------------------------------------+
    | placeholder       | description                                  |
    +-------------------+----------------------------------------------+
    | %(sql)s           | Formatted SQL statement that was executed    |
    +-------------------+----------------------------------------------+
    | %(sql_duration)s  | Formatted duration of SQL execution          |
    +-------------------+----------------------------------------------+

    Arguments:
        colorize_queries(bool): Should apply shell coloring escape sequences to
            formatted SQL?
        multiline_queries(bool): Should emit SQL as indented, multiline of
            single line log statements? In development it is usually nice to
            have it be `True`. In production environments, multiline logs are
            pain and should be avoided.

    Example::

        import logging
        from logging.config import dictConfig

        from seveno_pyutil import SQLFilter

        try:
            import sqlalchemy
            SQLFilter.register_sqlalchemy_logging_events('myapp.db')
        except ImportError:
            pass

        dictConfig({
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'django_sql': {'format': '(%(sql_duration)s) %(sql)s'},
                'sqlalchemy_sql': {'format': '(%(sql_duration)s) %(sql)s %(message)s'}
            },
            'filters': {
                'colored_sql': {
                    '()': 'seveno_pyutil.SQLFilter'
                    'colorize_queries': True,
                    'multiline_queries': True,
                }
            },
            'handlers': {
                'console_django': {
                    'class': 'logging.StreamHandler',
                    'level': 'DEBUG',
                    'formatter': 'django_sql',
                    'filters': ['colored_sql'],
                    'stream': 'ext://sys.stdout'
                },
                'console_sqlalchemy': {
                    'class': 'logging.StreamHandler',
                    'level': 'DEBUG',
                    'formatter': 'sqlalchemy_sql',
                    'filters': ['colored_sql'],
                    'stream': 'ext://sys.stdout'
                }
            },
            'loggers': {
                'myapp.db': {
                    'level': 'DEBUG',
                    'propagate': False,
                    'handlers': ['console_sqlalchemy']
                },
                'django.db.backends': {
                    'level': 'DEBUG',
                    'propagate': False,
                    'handlers': ['console_django']
                },
                'django.db.backends.schema': {
                    'level': 'DEBUG',
                    'propagate': False,
                    'handlers': ['console_django']
                }
            }
        })

    """

    KEY_SQL_CUMULATIVE_DURATION = "cumulative_duration"
    KEY_SQL_COUNT = "statements_count"

    @classmethod
    def register_sqlalchemy_logging_events(
        cls, logger, duration_threshold_ms=None, statistics_ctx_get=None
    ):
        """
        Must be called when logging SQLAlchemy statements and durations.

        Arguments:
            logger: name of logger or logging.Logger instance that will be used to log
                SQL messages.
            duration_threshold_ms: If given, only queries lasting longer than this
                threshold will be logged.
            statistics_ctx_get: callable that returns mutable dict. If given, then each
                time SQL query is executes, this dict will be updated.

                It can be used for example in Flask to track SQL execution statistics
                during one HTTP request::

                    import flask
                    from seveno_pyutil import register_sqlalchemy_logging_events

                    def sqlalchemy_stats():
                        return flask.g.setdefault("sqlalchemy_statistics", {})

                    register_sqlalchemy_logging_events(
                        "may_app_sql_logger", statistics_ctx_get=sqlalchemy_stats
                    )

                    @flask.before_request
                    def track_sqlalchemy():
                        @flask.after_this_request
                        def log_sqlalchemy_stats(response):
                            logger.info(
                                "SQLAlchemy statistics for request %s",
                                sqlalchemy_stats()
                            )
                            flask.g["sqlalchemy_statistics"] = {}
        """

        from sqlalchemy import event
        from sqlalchemy.engine import Engine

        _logger = logger
        if isinstance(logger, str):
            _logger = logging.getLogger(logger)

        if duration_threshold_ms is not None and not isinstance(
            duration_threshold_ms, timedelta
        ):
            duration_threshold_ms = timedelta(milliseconds=duration_threshold_ms)

        def duration_ms(conn):
            return timedelta(
                milliseconds=(
                    timeit.default_timer() - conn.info["query_start_time"].pop(-1)
                )
                * 1000.0
            )

        @event.listens_for(Engine, "before_cursor_execute")
        def before_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
        ):
            conn.info.setdefault("query_start_time", []).append(timeit.default_timer())

        @event.listens_for(Engine, "after_cursor_execute")
        def after_cursor_execute(
            conn, cursor, statement, parameters, context, executemany
        ):
            statement_duration = duration_ms(conn)

            warned = False
            if (
                duration_threshold_ms is not None
                and statement_duration >= duration_threshold_ms
            ):
                _logger.warning(
                    "Detected SQL query running longer than {:.2f} ms! ".format(
                        duration_threshold_ms.total_seconds() * 1000
                    ),
                    extra={
                        "sql_statement": str(statement),
                        "sql_parameters": str(parameters),
                        "sql_duration_ms": statement_duration.total_seconds() * 1000,
                    },
                )
                warned = True

            if not warned and statement_duration is not None:
                _logger.debug(
                    "",
                    extra={
                        "sql_statement": str(statement),
                        "sql_parameters": str(parameters),
                        "sql_duration_ms": statement_duration.total_seconds() * 1000,
                    },
                )

            if statistics_ctx_get:
                try:
                    data = statistics_ctx_get()
                    data.setdefault(cls.KEY_SQL_CUMULATIVE_DURATION, timedelta())
                    data.setdefault(cls.KEY_SQL_COUNT, 0)
                    data[cls.KEY_SQL_CUMULATIVE_DURATION] += statement_duration
                    data[cls.KEY_SQL_COUNT] += 1

                except Exception:
                    _logger.error("Failed to collect SQL statistics!", exc_info=True)

        @event.listens_for(Engine, "handle_error")
        def receive_handle_error(exception_context):
            _logger.critical(
                "",
                extra={
                    "sql_statement": str(exception_context.statement),
                    "sql_parameters": str(exception_context.parameters),
                    "sql_duration_ms": duration_ms(
                        exception_context.connection
                    ).total_seconds()
                    * 1000,
                },
            )

    def __init__(
        self, colorize_queries=False, multiline_queries=False, *args, **kwargs
    ):
        self.colorize_queries = colorize_queries
        self.multiline_queries = multiline_queries
        super().__init__(*args, **kwargs)

    def filter(self, record):
        sql = (
            # SQLAlchemy
            getattr(record, "sql_statement", None)
            # Django
            or getattr(record, "sql", None)
            or ""
        )

        if sql:
            if self.multiline_queries:
                sql = sqlparse.format(sql, reindent=True, keyword_case="upper").strip()
            else:
                sql = " ".join(
                    l.strip()
                    for l in sqlparse.format(
                        sql, reindent=True, keyword_case="upper"
                    ).splitlines()
                ).strip()

            if sql and not sql.endswith(";"):
                sql = sql + ";"

        if hasattr(record, "sql_parameters"):
            params_dict = record.sql_parameters
            params = json.dumps(record.sql_parameters, cls=JsonEncoder).strip()
        else:
            params_dict = {}
            params = ""

        if self.colorize_queries:
            if sql:
                sql = pygments.highlight(
                    sql, SqlLexer(), Terminal256Formatter(style="monokai")
                ).strip()

            if params:
                params = pygments.highlight(
                    params,
                    pygments.lexers.get_lexer_for_mimetype("application/json"),
                    Terminal256Formatter(style="monokai"),
                ).strip()

        # rsyslogd limits to 2048 bytes per message by default
        # We prefer to lose some params when shortening log line than to lose some SQL
        # statement content. Either way, it is not possible to guarnatee both will
        # be fully logged in all contexts and log sinks
        if params and params_dict:
            record.sql = "{} with params {}".format(sql[:1300], params[:500])
        else:
            record.sql = (sql or "SQL")[:1300]

        duration = getattr(record, "sql_duration_ms", None) or getattr(
            record, "duration", None
        )

        if duration:
            record.sql_duration = "{:.2f} ms".format(duration)
        else:
            record.sql_duration = "_.___ ms"

        return super().filter(record)



import logging
import socket
from datetime import datetime
from zoneinfo import ZoneInfo

import tzlocal


class StandardMetadataFilter(logging.Filter):
    """
    Filter that adds few more attributes to log records.

    +-------------------+----------------------------------------------+
    | placeholder       | description                                  |
    +-------------------+----------------------------------------------+
    | %(hostname)s      | hostname                                     |
    +-------------------+----------------------------------------------+
    | %(isotime)s       | Local time represented as ISO8601            |
    +-------------------+----------------------------------------------+
    | %(isotime_utc)s   | local time converted to UTC and represented  |
    |                   | as ISO8601 string                            |
    +-------------------+----------------------------------------------+
    """

    try:
        _HOSTNAME = socket.gethostname()
    except Exception as exception:
        _HOSTNAME = "-"

    _LOCAL_TZ = ZoneInfo(tzlocal.get_localzone_name())

    def filter(self, record):
        # Note that naive datetime will not be exported correctly to isoformat:
        #   datetime.now().isoformat()
        #   '2017-05-22T13:51:49.336335'
        # but we don't have this problem because we make sure that here we
        # always have tzinfo on datetime object
        dt = datetime.fromtimestamp(record.created).replace(tzinfo=self._LOCAL_TZ)

        record.isotime = dt.isoformat()
        record.isotime_utc = dt.astimezone(ZoneInfo("UTC")).isoformat()
        record.hostname = self._HOSTNAME

        return super().filter(record)



from seveno_pyutil import in_batches


class DescribeInBatches:
    def it_iterates_iterator_in_batches(self):
        g = (o for o in range(10))
        result = []
        for batch in in_batches(g, of_size=3):
            result.append(list(batch))

        assert result == [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

    def it_iterates_list_in_batches(self):
        g = list(range(10))
        result = []
        for batch in in_batches(g, of_size=3):
            result.append(list(batch))

        assert result == [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]

    def it_changes_resulting_batch_size_when_individual_batches_are_not_fully_consumed(
        self,
    ):
        g = list(range(10))
        result = []
        for batch in in_batches(g, of_size=3):
            result.append([next(batch), next(batch)])

        assert result == [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]



import itertools


def in_batches(iterable, of_size=1):
    """
    Generator that yields generator slices of iterable.

    Since it is elegant and working flawlessly, it is shameles C/P from
    https://stackoverflow.com/questions/8991506/iterate-an-iterator-by-chunks-of-n-in-python/8998040#8998040

    Warning:
        Each returned batch should be completely consumed before next batch
        is yielded. See example below to better understand what that means.

    Example::

        from seveno_pyutil import in_batches

        g = (o for o in range(10))
        for batch in in_batches(g, of_size=3):
            print(list(batch))
        # [0, 1, 2]
        # [3, 4, 5]
        # [6, 7, 8]
        # [9]

        # And don't consume whole batch before yielding another one...
        g = list(range(10))
        for batch in in_batches(g, of_size=3):
            print( [next(batch), next(batch)] )
        # [0, 1]
        # [2, 3]
        # [4, 5]
        # [6, 7]
        # [8, 9]
    """
    it = iter(iterable)
    while True:
        chunk_it = itertools.islice(it, of_size)
        try:
            first_el = next(chunk_it)
        except StopIteration:
            return
        yield itertools.chain((first_el,), chunk_it)



import sys
from collections import abc
from importlib import import_module
from typing import Mapping, Union


class LeafSubclassRetriever:
    """
    http://code.activestate.com/recipes/577858-concrete-class-finder/
    """

    def __init__(self, base_class):
        self.base_class = base_class

    def value(self):
        direct_subclasses = self.base_class.__subclasses__()
        leaf_subclasses = list()
        for klass in direct_subclasses:
            if len(klass.__subclasses__()) > 0:
                leaf_subclasses += LeafSubclassRetriever(klass).value()
            else:
                leaf_subclasses.append(klass)

        return leaf_subclasses


def leaf_subclasses(klass):
    """
    Returns all leaf subclasses of given ``klass``
    """
    return LeafSubclassRetriever(klass).value()


def all_subclasses(klass):
    subclasses = set(klass.__subclasses__())

    for subklass in subclasses:
        subclasses = subclasses.union(all_subclasses(subklass))

    return subclasses


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by
    the last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as e:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from e

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as e:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from e


def getval(src: Union[Mapping, object], attr: object, default=None):
    """
    Companion of `dict.get` and `getattr` which ensures default value even when
    original method would had returned `None`

    Example::

        d1 = {}
        d2 = {"foo": None}
        d3 = {"foo": "bar"}

        d1.get("foo", 42)  # => 42
        d2.get("foo", 42)  # => None
        d3.get("foo", 42)  # => "bar"

        getval(d1, "foo", 42)  # => 42
        getval(d2, "foo", 42)  # => 42
        getval(d3, "foo", 42)  # => "bar"

        # and also

        getval({"a": ""}, "a", None) is None  # => True

        # and with objects other than dicts

        class Foo:
            def __init__(self, foo=None):
                self.foo = foo

        class Bar:
            pass

        o1 = Bar()
        o2 = Foo()
        o3 = Foo("bar")

        getattr(o1, "foo", 42)  # => 42
        getattr(o2, "foo", 42)  # => None
        getattr(o3, "foo", 42)  # => "bar"

        getval(o1, "foo", 42)  # => 42
        getval(o2, "foo", 42)  # => 42
        getval(o3, "foo", 42)  # => "bar"
    """

    if isinstance(src, abc.Mapping):
        return src.get(attr, None) or default
    else:
        return getattr(src, attr, None) or default
